#ifndef OVERLAP_AREA
#define OVERLAP_AREA

double computeOverlap(double *ilon, double *ilat, double *olon, double *olat,
                      int energyMode, double refArea, double *areaRatio);

#endif
